from __future__ import print_function

'''
   Wave Viewer is a video viewer modifiied to view
   a wave field expreesed by complex number.

   figobj which has "set_phasor" can dynamically
   compute a new data set using phasor.

   (Presently, only fig_tripcolor has this feature.)

   Usage:
         from ifigure.widgets.wave_viewer import WaveViewer

         v = figure(viewer = WaveViewer, nframe = 15, absolute = True)

             nframe: number of frame
             absolute: if true, the last frame is absolute 
'''
import wx
from ifigure.widgets.video_viewer import VideoBookPlayer, VideoViewerMode
from ifigure.widgets.book_viewer import BookViewer, BookViewerFrame
from ifigure.widgets.book_viewer_interactive import allow_interactive_call
import ifigure.events
import numpy as np
import weakref

from ifigure.mto.fig_tripcolor import FigTripcolor
from ifigure.mto.fig_quiver import FigQuiver
from ifigure.mto.fig_plot import FigPlot
from ifigure.mto.fig_solid import FigSolid
from ifigure.mto.fig_surface import FigSurface
from ifigure.mto.fig_image import FigImage
from matplotlib.collections import TriMesh


class FigTripcolorPhasor(FigTripcolor):
    def set_phasor(self, angle=None):
        tri = self.get_masked_tri()
        if angle is not None:
            z = self.getp('z') * np.exp(1j*angle)
            if isinstance(self._artists[0], TriMesh):
                C = z.flatten().real
            else:
                C = (z.flatten().real)[tri].mean(axis=1)
        else:
            z = np.absolute(self.getp('z')).flatten()
            if isinstance(self._artists[0], TriMesh):
                C = z.flatten()
            else:
                C = z[tri].mean(axis=1)
        self._artists[0].set_array(C)

class FigPlotPhasor(FigPlot):
    def set_phasor(self, angle=None):
        y = self._artists[0].get_ydata()
        y_complex = self.getvar('complex_y')
        if angle is not None:
            new_y = (y_complex * np.exp(1j*angle))[:len(y)]
            new_y = new_y.real
        else:
            new_y = np.absolute(y_complex[:len(y)])
        self._artists[0].set_ydata(new_y)
        self.setp('y', new_y)

class FigQuiverPhasor(FigQuiver):
    def set_phasor(self, angle=None):
        u = self.getvar('complex_u')
        v = self.getvar('complex_v')
        y_complex = self.getvar('y')
        if angle is not None:
            self.setvar('u', (u * np.exp(1j*angle)).real)
            self.setvar('v', (v * np.exp(1j*angle)).real)
        else:
            self.setvar('u', np.absolute(u))
            self.setvar('v', np.absolute(v))
        self.reset_artist()


class FigSolidPhasor(FigSolid):
    def set_phasor(self, angle=None):
        if angle is not None:
            z = self.getvar('cdata') * np.exp(1j*angle)
#           if len(z.shape) == 2:
#              z =  np.mean(z, -1).real
#           else:
            z = z.real
        else:
            z = np.absolute(self.getvar('cdata'))
#           if len(z.shape) == 2:
#              z =  np.mean(z, -1)
        self._artists[0]._gl_facecolordata = z
        self._artists[0]._update_fc = True


class FigImagePhasor(FigImage):
    def set_phasor(self, angle=None):
        z = self.getvar('complex_z')
        if angle is not None:
            self.setvar('z', (z * np.exp(1j*angle)).real)
        else:
            self.setvar('z', np.absolute(z))
        self.reset_artist()

    def get_crange(self, crange=[None, None],
                   xrange=[None, None],
                   yrange=[None, None],
                   scale='linear'):
        z = self.getvar('complex_z')        
        return [-np.max(np.abs(z)), np.max(np.abs(z))]
    
    def get_export_val(self, a):
        x, y, z = self._eval_xyz()
        z = self.getvar('complex_z')
        return {"cdata": z,
                "xdata": x,
                "ydata": y}

class FigSurfacePhasor(FigSurface):
    def set_phasor(self, angle=None):
        if angle is not None:
            z = self.getvar('cdata') * np.exp(1j*angle)
#           if len(z.shape) == 2:
#              z =  np.mean(z, -1).real
#           else:
            z = z.real
        else:
            z = np.absolute(self.getvar('cdata'))
#           if len(z.shape) == 2:
#              z =  np.mean(z, -1)
        r, c, idxset = self._artists[0]._idxset
        if r is not None:
            # surface plot
            fc = z[r, :][:, c].flatten()[idxset]
        else:
            # trisurf plot
            fc = z.flatten()
        if idxset is not None:
            fc = fc[idxset]
        self._artists[0]._gl_facecolordata = fc
        self._artists[0]._update_fc = True


def convert_figobj(obj):
    if obj.__class__ == FigTripcolor:
        obj.__class__ = FigTripcolorPhasor
    elif obj.__class__ == FigPlot:
        obj.__class__ = FigPlotPhasor
    elif obj.__class__ == FigQuiver:
        obj.__class__ = FigQuiverPhasor
    elif obj.__class__ == FigSolid:
        if obj.getvar('cdata') is None:
            return
        obj.__class__ = FigSolidPhasor
    elif obj.__class__ == FigSurface:
        if obj.getvar('cdata') is None:
            return
        obj.__class__ = FigSurfacePhasor
    elif obj.__class__ == FigImage:
        obj.__class__ = FigImagePhasor
    else:
        pass


class WaveViewer(VideoBookPlayer):
    def __init__(self, *args, **kwargs):
        self.nframe = kwargs.pop('nframe', 30)
        self.sign = kwargs.pop('sign',  -1)
        super(WaveViewer, self).__init__(*args, **kwargs)

        if self.book is not None:
            self.add_all_video_obj()

    def add_bookmenus(self, editmenu, viewmenu):
        VideoViewerMode.add_bookmenus(self, editmenu, viewmenu)
        self.add_menu(viewmenu, wx.ID_ANY,
                      "Show Absolute",  "show absolute value of wave field",
                      self.onShowAbs)

    def tripcolor(self, *args, **kwargs):
        o = BookViewer.tripcolor(self, *args, **kwargs)
        if o is None:
            return
        if o.getvar('z').dtype.name.startswith('complex'):
            convert_figobj(o)
            self.add_video_obj(o)
        return o

    def plot(self, *args, **kwargs):
        o = BookViewer.plot(self, *args, **kwargs)
        if o is None:
            return
        y = args[1]
        if y.dtype.name.startswith('complex'):
            o.setvar('complex_y', y)
            convert_figobj(o)
            self.add_video_obj(o)
        return o

    @allow_interactive_call
    def quiver(self, *args, **kwargs):
        try:
            o = FigQuiver(*args, **kwargs)
        except ValueError as x:
            print(x.message)
            return
        if o.getvar('u').dtype.name.startswith('complex'):
            convert_figobj(o)
            self.add_video_obj(o)
            o.setvar('complex_u', o.getvar('u'))
            o.setvar('complex_v', o.
                     getvar('v'))
            o.setvar('u', o.getvar('u').real)
            o.setvar('v', o.getvar('v').real)
        return o

    # @allow_interactive_call
    def solid(self, *args, **kwargs):
        try:
            o = BookViewer.solid(self, *args, **kwargs)
        except ValueError as x:
            return
        convert_figobj(o)
        self.add_video_obj(o)
        return o

    def surf(self, *args, **kwargs):
        try:
            o = BookViewer.surf(self, *args, **kwargs)
        except ValueError as x:
            return
        convert_figobj(o)
        self.add_video_obj(o)
        return o

    def image(self, *args, **kwargs):
        if len(args) == 1:
            complex_z = args[0]
            if np.iscomplexobj(complex_z):
               args = (complex_z.real,)
        elif len(args) == 3:
            complex_z = args[2]
            if np.iscomplexobj(complex_z):
               args = (args[0], args[1], complex_z.real,)
        else:
            print("incorrect number of arguments: image(z) or image(x, y, z)")
            return
        
        try:
            o = BookViewer.image(self, *args, **kwargs)
        except ValueError as x:
            return
        o.setvar('complex_z', complex_z)                
        convert_figobj(o)
        self.add_video_obj(o)
        return o
        
    def _get_phase(self, ipage):
        return self.sign*np.pi*2*ipage/self.nframe

    def num_page(self):
        return self.nframe

    def add_all_video_obj(self):
        self.reset_video_obj_set()
        for obj in self.book.walk_tree():
            if isinstance(obj, FigTripcolorPhasor):
                self.add_video_obj(obj)
            elif isinstance(obj, FigPlotPhasor):
                self.add_video_obj(obj)
            elif isinstance(obj, FigQuiverPhasor):
                self.add_video_obj(obj)
            elif isinstance(obj, FigSolidPhasor):
                self.add_video_obj(obj)
            elif isinstance(obj, FigSurfacePhasor):
                self.add_video_obj(obj)
            elif isinstance(obj, FigImagePhasor):
                self.add_video_obj(obj)
            else:
                pass

    def UpdateImage(self, i):
        phase = self._get_phase(i)
        for obj in self._video_obj:
            obj.set_phasor(phase)
            obj.set_bmp_update(False)
        self._video_page = i
        self.draw()

    def onShowAbs(self, evs):
        for obj in self._video_obj:
            obj.set_phasor(None)
            obj.set_bmp_update(False)
        self.draw()

    def onNextVideoPage(self, evt):
        if self._video_page == self.nframe - 1:
            return
#        self._video_page = self._video_page + 1
        ifigure.events.SendShowPageEvent(self.book, self, '1')

    def onPrevVideoPage(self, evt):
        if self._video_page == 0:
            return
#        self._video_page = self._video_page - 1
        ifigure.events.SendShowPageEvent(self.book, self, '-1')

    def videoviewer_config(self):
        from ifigure.utils.edit_list import DialogEditList
        l = [
            ["Interval(sec.)", str(float(self._playinterval)/1000.),
             0, {'noexpand': True}],
            ["Frame #", str(self.nframe),
             0, {'noexpand': True}],
            [None, self._playloop, 3, {
                "text": "Loop Movie", "noindent": None}],
            [None, self.sign == -1, 3,
                {"text": "Use exp(-iwt)", "noindent": None}],
        ]
        value = DialogEditList(l, parent=self,
                               title='Player Config.',
                               style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER)
        if not value[0]:
            return
        self._playinterval = int(float(value[1][0])*1000.)
        self._playloop = bool(value[1][2])
        self.sign = -1 if bool(value[1][3]) else 1
        self.nframe = int(value[1][1])



def piscope():
    import multiprocessing as mp
    import mpl_toolkits
    import sys
    import os
    import wx
    import weakref
    import matplotlib
    import shutil
    import warnings

    import platform
    if platform.system() == 'Darwin':
        mp.set_start_method('spawn')
    else:
        if 'forkserver' in mp.get_all_start_methods():
            mp.set_start_method('forkserver')

    matplotlib.use('WXAGG')

    wx.UpdateUIEvent.SetMode(wx.UPDATE_UI_PROCESS_SPECIFIED)
    wx.UpdateUIEvent.SetUpdateInterval(1000)

    import ifigure
    from ifigure.ifigure_app import ifigure_app, MyApp
    from ifigure.utils.mp_tarzip import MPTarzip
    from os.path import expanduser

    # if it does not have write permission to current directory
    # it moves to home directory
    # make tempfile need write permission in some case?
    # (2014 10)
    if not os.access(os.getcwd(), os.W_OK):
        print('No access to current working directory, moving to home directory')
        os.chdir(expanduser('~'))

    import site
    home = expanduser('~')
    site.USER_SITE = os.path.join(
        home, '.ifigure_rc', '.local', 'site-packages')
    site.USER_BASE = os.path.join(home, '.ifigure_rc', '.local')
    redirect_std = True
    use_console = True
    file = None
    start_server = False
    show_file_open_error = False
    exe_command = None
    hide_main = False

    class MainlevelJob(object):
        def __init__(self):
            self.func = None
            self.func2 = []

        def setjob(self, func):
            self.func = func

        def dojob(self):
            if self.func is not None:
                self.func()
            self.func = None

        def setfinishjob(self, func):
            if not func in self.func2:
                self.func2.append(func)

        def finishjob(self):
            for f in self.func2:
                f()

    # this is a place where wdir is set when exiting
    # the program
    import __main__
    __main__.xxx = []
    launcher_file = None
    use_gl = True
    if len(sys.argv[1:]) >= 1:
        rflag = False
        lflag = False
        eflag = False
        for p in sys.argv[1:]:
            if p == '-h':
                print('[Usage: ifigure -s -r command -h file]')
                print('ifigure          : start a new project')
                print('ifigure <file>   : open an existing project')
                print('-s               : start server thread')
                print('-d               : suppress console redirect')
                print('-c               : completely suppress redirect')
                print('-n               : no main window')
                print('-p               : call profiler')
                print('-r <command>     : run command')
                print('-h               : show this help')
                print('-g               : turn off gl')
                print('-w               : warning on')
                print('-ww              : warning on (error)')
                print('-l <path>        : file to commnicate with launcher ')
                sys.exit()
            elif p == '-s':
                start_server = True
                redirect_std = True
                __main__.process_server_request = False

                server = ifigure.server.Server()
                server.start()
                continue
            elif p == '-e': ## this is backword compatibility (ignored)
                eflag = True
            elif p == '-d':
                use_console = False
                redirect_std = False
                print('debug mode (console is suppressed. all redirect is off')
                continue
            elif p == '-c':
                use_console = False
                print('debug mode (console is suppressed, but thread log  window is on)')
                continue
            elif p == '-n':
                hide_main = True
                continue
            elif p == '-p':
                pr = None
                import cProfile
                print('starting profiler')
                pr = cProfile.Profile()
                pr.enable()
            elif p == '-r':
                rflag = True
            elif p == '-l':
                lflag = True
            elif p == '-g':
                use_gl = False
            elif p == '-w':
                warnings.simplefilter('always')
                print('debug mode (warning is on)')
            elif p == '-ww':
                warnings.simplefilter('error')
                print('debug mode (warning cuases error)')
            else:
                if rflag:
                    if len(p) > 0:
                        p = p.strip()
                        if p.startswith('"'):
                            exe_command = p[1:-1]
                        elif p.startswith("'"):
                            exe_command = p[1:-1]
                        else:
                            exe_command = p
                    if exe_command.strip() == '':
                        exe_command = None
                    rflag = False
                elif lflag:
                    launcher_file = p.strip()
                    lflag = False
                elif eflag:
                    eflag = False
                else:
                    if os.path.exists(p):
                        file = p
                        print(('opening file : '+file))
                        file = os.path.abspath(file)
                    else:
                        show_file_open_error = True
                        filename = p
                        file = None
    from ifigure.widgets.appearance_config import AppearanceConfig
    appearanceconfig = AppearanceConfig()

    import ifigure.widgets.canvas.ifigure_canvas
    import ifigure.matplotlib_mod.backend_wxagg_gl as wxagg_gl
    if use_gl:
        use_gl = appearanceconfig.setting['gl_use']
    if use_gl:
        ifigure.widgets.canvas.ifigure_canvas.turn_on_gl = use_gl
        wxagg_gl.use_gl_12 = appearanceconfig.setting['gl_use_12']
        wxagg_gl.load_glcanvas(debug=True)
    else:
        print('No 3D plot (OpenGL turned off)')

    ifigure.ifigure_app.redirect_std = redirect_std
    ifigure.ifigure_app.use_console = use_console
#   from ifigure.utils.rollback_importer import RollbackImporter as RI
    from ifigure.mto.treedict import fill_td_name_space

    sc = os.path.join(os.path.dirname(ifigure.__file__),
                      'mto', 'treedict_ns.py')
    if os.path.exists(sc):
        fill_td_name_space(sc)
    from ifigure.ifigure_config import rcdir
    sc = os.path.join(rcdir, 'treedict_ns.py')
    if os.path.exists(sc):
        fill_td_name_space(sc)

    app = MyApp(False, clearSigInt=False)
    ifig_app = app.get_ifig_app()

    if show_file_open_error:
        ifig_app.shell.write('### File not found : ' + filename)

    if file is not None:
        if file[-4:] == '.pfz':
            ifig_app.proj_tree_viewer.update_widget()
            #ifig_app.open_file(file, call_close=True)
            # ifig_app.set_proj_saved(True)
            #wx.CallAfter(ifig_app.onOpen, path =file)
            # somehow this seems work, but others may not open
            # figure windows associated to the project file
            wx.CallLater(10, ifig_app.onOpen, path=file)

            #           ifig_app.draw_all()
            # ifig_app.set_filename_2_window_title()
        elif file[-4:] == '.bfz':
            bk = ifig_app.proj.load_subtree(file, compress=True)
            from ifigure.mto.fig_book import FigBook
            if not isinstance(bk, FigBook):
                sys.exit()
            ifig_app.ipage = 0
            bk.setvar("original_filename", file)
#           bk.set_open(True)
            ifig_app.book = bk
#           obk.destroy()
            ifigure.events.SendChangedEvent(bk, w=ifig_app)
            ifigure.events.SendOpenBookEvent(bk, w=ifig_app)
#           ifig_app.show_page(ifig_app.ipage)
        else:
            wx.CallLater(3, ifig_app.onOpenWithHelperCommand,
                         path=file, hide_main=hide_main)
#       ifig_app.open_book_in_appwindow(ifig_app.proj.book1, ipage=0)

    if start_server:
        ifig_app.use_server()
        __main__.process_server_request = True
        port = server.info()[3]
        print('remote port is open : port = ' + str(port) + '\n')

    # call tempdir_clean when ifig_app is being deleted
    class TempdirObj(object):
        pass
    ifig_app._tempdir_obj = TempdirObj()
    from ifigure.ifigure_config import tempdir_clean
    tempdir_ref = weakref.ref(ifig_app._tempdir_obj, tempdir_clean)

    # reduce update events
    wx.UpdateUIEvent.SetMode(wx.UPDATE_UI_PROCESS_SPECIFIED)

#   if pr is not None:
#      from ifigure.interactive import profile_stop
#      profile_stop(pr, sortby='cumulative')

    if exe_command is not None:
        if hide_main:
            wx.CallAfter(ifig_app.shell.execute_and_hide_main, exe_command)
        else:
            wx.CallAfter(ifig_app.shell.Execute, exe_command)
    if hide_main and exe_command is None:
        # i don't know if this is necessary hide_main and exe_command
        # is used together for normal situation.
        wx.CallAfter(ifig_app.goto_no_mainwindow)
    # conditions for iptyhon
    ifig_app.set_launcher_file(launcher_file)

    from ifigure.widgets.taskbar import TaskBarIcon
    tbicon = TaskBarIcon()

    app.MainLoop()

    server = ifigure.server.Server()
    if server.info()[0]:
        server.stop()

    if not MPTarzip().isReady():
        # seems like it is not necessary since wx.CallLater
        # im MPTarzip makes sure that the program does not
        # come here before save process finishes.

        # anyway, just in case...
        print('waiting for save to be done')
        MPTarzip().worker.join()
    #
    #  deleting the wdir used last moment...
    #
    wdirs = __main__.xxx
    for wdir in wdirs:
        if os.path.exists(wdir):
            print(('deleting :', wdir))
            shutil.rmtree(wdir)
#   MDSWorkerPool(type=worker_mode).reset()
    print('#### Exiting piScope (main loop finished)')
    from .ifigure_config import tempdir
    if os.path.exists(tempdir) and len(os.listdir(tempdir))==0:
        print(' --- removing workdirectory :' + tempdir)
        os.rmdir(tempdir)
    print(' --- ebug to check if normal exit (list of remaining threads)')
    import threading
    import time
    time.sleep(1)
    for t in threading.enumerate():
        print(" --- "+str(t))
    print(" --- "+str(wx.GetTopLevelWindows()))



from __future__ import print_function

__author__ = "Syun'ichi Shiraiwa"
__copyright__ = "Copyright, S. Shiraiwa, PiScope Project"
__credits__ = ["Syun'ichi Shiraiwa"]
__version__ = "0.5"
__maintainer__ = "S. Shiraiwa"
__email__ = "shiraiwa@psfc.mit.edu"
__status__ = "beta"


from ifigure.mto.fig_obj import FigObj
from ifigure.mto.fig_grp import FigGrp
from ifigure.mto.fig_page import FigPage
from ifigure.widgets.canvas.file_structure import *
import ifigure
import os
import ifigure.utils.cbook as cbook
import ifigure.utils.geom as geom
import ifigure.widgets.canvas.custom_picker as cpicker
import numpy as np
import weakref
import logging
import matplotlib
from matplotlib.patches import Rectangle
from matplotlib.legend import Legend
import matplotlib.path as mpath
import matplotlib.cbook as mplcbook
import matplotlib.transforms as mpltransforms

from ifigure.widgets.undo_redo_history import GlobalHistory
from ifigure.widgets.undo_redo_history import UndoRedoArtistProperty
from ifigure.widgets.undo_redo_history import UndoRedoFigobjProperty
from ifigure.widgets.undo_redo_history import UndoRedoFigobjMethod

from ifigure.mto.generic_points import GenericPoint, GenericPointsHolder
from ifigure.mto.figobj_gpholder import FigObjGPHolder


class FigLegend(FigObjGPHolder):
    def __init__(self, *args, **kywds):
        '''
        Curve object:
           figaxe1, figaxes2: None or full_path string
        '''

        self._objs = []  # for debug....
        if "draggable" in kywds:
            self.setvar("draggable", kywds["draggable"])
            del kywds["draggable"]
        else:
            self.setvar("draggable", False)
        if "container_idx" in kywds:
            self._container_idx = kywds["container_idx"]
            del kywds["container_idx"]
        else:
            self._container_idx = 0

        kywds = self._process_kywd(kywds, "xy", [0, 0])
        kywds = self._process_kywd(kywds, "trans1", ['figure']*2)
        kywds = self._process_kywd(kywds, "figaxes1", None)
        kywds = self._process_kywd(kywds, "fbedgecolor", 'k')
        kywds = self._process_kywd(kywds, "fbfacecolor", 'w')
        kywds = self._process_kywd(kywds, "fblinewidth", 1.0)
        kywds = self._process_kywd(kywds, "fblinestyle", "solid")
        kywds = self._process_kywd(kywds, "fbalpha", 1.0)
        kywds = self._process_kywd(kywds, "title", None)
        kywds = self._process_kywd(kywds, "shadow", False)

        GenericPointsHolder.__init__(self, num=1)

        if len(args) == 1:
            self.setvar('legendlabel', args[0])
        elif len(args) == 2:
            if (mplcbook.is_string_like(args[1]) or
                    isinstance(args[1], int)):
                self.setvar('legendlabel', args[0])
                self.setvar('legendloc', args[1])
            else:
                self.setvar('legendhandle', args[0])
                self.setvar('legendlabel', args[1])
        args = ()

        super(FigLegend, self).__init__(*args, **kywds)

        if self.getvar('figaxes1') is not None:
            self.setvar("figaxes1", figaxes1.get_full_path())

        self.setp('use_var', True)
        self.setp('legendpos', (None, None))
        self.setp('zorder', 2.5)
        self.setp('legendlabelprop', ['default']*4)

        self._cb_added = False
        self._drag_mode = 0
        self._drag_artist = None
        self._legend_hitlevel = -1
        self._hit_a = None
        self._floating = True

    @classmethod
    def isFigLegend(self):
        return True

    @classmethod
    def can_have_child(self, child=None):
        return False

    @classmethod
    def get_namebase(self):
        return 'legend'

    @classmethod
    def allow_outside(self):
        ### if fig_obj can be dragged outside of axes ###
        return True

    @classmethod
    def property_in_file(self):
        return []

    @classmethod  # define _attr values to be saved
    def attr_in_file(self):
        return (["legendlabelcolor", "legendcolor", "legendlabel",
                 "legendlabelprop",
                 "legendpos", "legendloc", "draggable"] +
                super(FigLegend, self).attr_in_file())

    @classmethod
    def property_in_palette(self):
        return ["position", "labels", "box"], [
            ["legendloc",  "legendlabelprop"],
            ["legendtitle", "legendentry"],
            ["facecolor_2", "edgecolor_2", "linewidthz",
             "plinestyle_2", "alpha_2", "legendshadow"],
        ]

    @classmethod
    def load_classimage(self):
        from ifigure.ifigure_config import icondir as path
        idx1 = cbook.LoadImageFile(path, 'legend.png')
        return [idx1]

    def set_parent(self, parent):
        FigObj.set_parent(self, parent)
        GenericPointsHolder.set_parent(self, parent)

    def canvas_menu(self):
        return []

    def get_artists_for_pick(self):
        return self._artists

    def isDraggable(self):
        if self.getp("draggable") is not None:
            return self.getp("draggable")
        else:
            return self.getvar("draggable")

    def generate_artist(self):
        if self.isempty() is False:
            return
        if self.getp('use_var'):
            if self.getvar('legendhandle') is not None:
                a = [weakref.ref(figobj._artists[0])
                     for figobj in self.getvar('legendhandle')]
                self.setp('legendhandle', a)
            if self.getvar('legendlabel') is not None:
                self.setp('legendlabel', self.getvar('legendlabel'))
            if self.getvar('legendloc') is not None:
                self.setp('legendloc', self.getvar('legendloc'))
            else:
                self.setp('legendloc', 'best')
            if self.getvar('draggable') is not None:
                self.setp('draggable', self.getvar('draggable'))
            else:
                self.setp('draggable', False)

            trans1 = self.getvar("trans1")
            figaxes1 = self.getvar("figaxes1")
            if figaxes1 is not None:
                figaxes1 = self.find_by_full_path(str(figaxes1))

            xy = self.getvar("xy")
            self.set_gp_point(0, xy[0], xy[1],
                              trans=trans1)
            self.set_gp_figpage()
            self.set_gp_figaxes(0, figaxes1)
            self.setp('use_var', False)

        a = self.make_newartist()
        if self.hasp("loaded_property"):
            self.set_artist_property(a, lp[0])
            self.delp("loaded_property")
        if not self._cb_added:
            fig_page = self.get_figpage()
            fig_page.add_resize_cb(self)
            self._cb_added = True

        self.setp('legendlabel', [t.get_text() for t in a.texts])
        return

    def onResize(self, evt):
        self.refresh_artist()

    def do_update_artist(self):
        self.refresh_artist()

    def apply_fbprop(self, patch):
        props = ['alpha', 'linestyle', 'linewidth',
                 'facecolor', 'edgecolor']
        for p in props:
            value = self.getvar('fb'+p)
            m = getattr(patch, 'set_'+p)
            m(value)

    def make_newartist(self):
        self.check_loaded_gp_data()
        c = self.get_container()
        args = []
        kwargs = self.getvar('kywds').copy()
        # clean legendhandle first
        if self.getp('legendhandle') is not None:
            v = [h for h in self.getp('legendhandle')
                 if (hasattr(h(), 'figobj') and
                     h().figobj is not None)]
            if len(v) == 0:
                self.delp('legendhandle')
            else:
                self.setp('legendhandle', v)

        if self.getp('legendhandle') is not None:
            v = [v() for v in self.getp('legendhandle')
                 if v() is not None]
            args.append(v)
        else:
            if hasattr(self.get_container(), "_get_legend_handles"):
                h = self.get_container()._get_legend_handles()
                args.append([x for x in h if hasattr(x, 'figobj')])
                self.setp('legendhandle', [weakref.ref(x)
                                           for x in h if hasattr(x, 'figobj')])

        if self.getp('legendlabel') is not None:
            args.append(self.getp('legendlabel'))
        if self.getp('legendloc') is not None:
            kwargs['loc'] = self.getp('legendloc')

        kwargs['title'] = self.getvar('title')
        kwargs['shadow'] = self.getvar('shadow')

        a = c.legend(*args, **kwargs)
        self.apply_fbprop(a.legendPatch)
        if self.getp('draggable'):
            pos = self.getp('legendpos')
            if pos[0] is not None:
                a._loc = tuple(self.getp('legendpos'))
        # store labelcolor, labelfamily, labelweight, labelstyle, labelsize
        labelcolor = self.getp('legendlabelcolor')
        if labelcolor is not None:
            i = 0
            while i < len(labelcolor) and i < len(a.texts):
                a.texts[i].set_color(labelcolor[i])
                i = i + 1
        labelcolor = [t.get_color() for t in a.texts]
        self.setp('legendlabelcolor', labelcolor)

        if self.get_figaxes() is not None:
            c.legend_ = None
            c.add_artist(a)

        if self.getp('legendhandle') is not None:
            for h in self.getp('legendhandle'):
                if (hasattr(h(), 'figobj') and
                        h().figobj is not None):
                    h().figobj.add_update_client(self)
                else:
                    print('h() does not have figobj')
#        xy = self.get_gp(0).get_device_point()

        lp = self.getp("loaded_property")
        if lp is not None:
            self.set_artist_property(a, lp[0])
            self.delp("loaded_property")
        a.figobj = self
        a.figobj_hl = []
        a.set_zorder(self.getp('zorder'))

        self._artists = [a]
        return a

    def refresh_artist(self):
        if len(self._artists) != 1:
            return
        a1 = self._artists[0]
        z = a1.zorder
        hl = len(self._artists[0].figobj_hl) != 0
        self.del_artist(delall=True)
        a = self.make_newartist()
        a.set_zorder(z)
        self.highlight_artist(hl, [a])
        a2 = self._artists[0]
        if a1 != a2:
            ifigure.events.SendReplaceEvent(self, a1=a1, a2=a2)

    def del_artist(self, artist=None, delall=False):
        if delall:
            artistlist = self._artists
        else:
            artistlist = artist

        # save_data2->load_data2 will set "loaded_property"
        self.load_data2(self.save_data2({}))
#        val = []
#        for a in self._artists:
#           val.append(self.get_artist_property(a))
#        if len(val) != 0:
#           self.setp("loaded_property", val)

        if len(artistlist) != 0:
            c = self.get_container()
            self.highlight_artist(False, artistlist)
            for a in artistlist:
                if self.get_figaxes() is None:
                    if a in c.legends:
                        c.legends.remove(a)
                else:
                    if a in c.artists:
                        a.remove()#c.artists.remove(a)

        super(FigLegend, self).del_artist(artistlist)

    def highlight_artist(self, val, artist=None):
        if artist is None:
            alist = self._artists
        else:
            alist = artist
        figure = self.get_figpage()._artists[0]
        if val == True:
            a = alist[0]
            if self._hit_a is None:
                box = a.get_window_extent().get_points()
            else:
                box = self._hit_a.get_window_extent().get_points()
            x = [box[0][0], box[0][0], box[1][0], box[1][0], box[0][0]]
            y = [box[0][1], box[1][1], box[1][1], box[0][1], box[0][1]]
            hl = matplotlib.lines.Line2D(x, y, marker='s',
                                         color='k', linestyle='None',
                                         markerfacecolor='none',
                                         markeredgewidth=0.5,
                                         figure=figure)
            figure.lines.extend([hl])
            a.figobj_hl.append(hl)
        else:
            for a in alist:
                for hl in a.figobj_hl:
                    figure.lines.remove(hl)
                    a.figobj_hl = []

    def get_artist_extent(self, a):
        #        return FigGrp.get_artist_extent(self, a)
        box = a.get_window_extent().get_points()
        #x0, x1, y0, y1
        return [box[0][0], box[1][0], box[0][1], box[1][1]]

    def get_artist_extent_all(self):
        return self.get_artist_extent(self._artists[0])

    def picker(self, a, evt):
        return self.picker_a(a, evt)

    def picker_a0(self, a, evt):
        member_hit = False
        self._picker_a_mode = 0
        self._hit_a = None
        hit, extra, type, loc = super(FigLegend, self).picker_a0(a, evt)

        for t in a.get_texts() + [self._artists[0].get_title()]:
            if t is None:
                continue
            try:
                box = self.get_artist_extent(t)
            except:
                continue
            if (evt.x > box[0] and
                evt.x < box[1] and
                evt.y > box[2] and
                    evt.y < box[3]):
                member_hit = True
                self._hit_a = t
                self.highlight_artist(False)
                break
        if hit:
            type = 'area'
            loc = 3
        return hit, extra, type, loc

    def drag_a(self, a, evt, shift=None, scale=None):
        redraw, scale = super(FigLegend, self).drag_a(
            a, evt, shift=shift, scale=scale)
        return redraw, scale

    def dragdone_a(self, a, evt, shift=None, scale=None):
        shift = evt.guiEvent_memory.ShiftDown()
        redraw, scale0 = super(FigLegend, self).dragdone_a(a, evt,
                                                           shift=shift, scale=scale)
        x = min(self._drag_rec[:2])
        y = min(self._drag_rec[2:])

        bbx = a.get_bbox_to_anchor()
        pos = mpltransforms.BboxTransformFrom(bbx).transform_point((x, y))

        h = [UndoRedoFigobjMethod(a, 'legendloc',
                                  (True, self.getp('legendloc'), pos))]

        window = evt.guiEvent_memory.GetEventObject().GetTopLevelParent()
        hist = GlobalHistory().get_history(window)
        self._hit_a = None
        self._picker_a_mode = 0
#        h = h + self.scale_artist(scale0, action = True)

        if len(h) != 0:
            hist.start_record()
            for item in h:
                hist.add_history(item)
            hist.stop_record()
        return 0, scale0

    def get_legendlabel(self, a, tab):
        idx = self._tab_2_idx(tab)
        if len(self._get_texts()) > idx:
            return self._get_texts()[idx].get_text()
        else:
            return ''

    def set_legendlabel(self, param, a):
        tab = param[0]
        value = param[1]
        idx = self._tab_2_idx(tab)
        if len(self._get_texts()) > idx:
            self._get_texts()[idx].set_text(value)
            h = self._artists[0].legendHandles[idx]
            h.set_label(value)
            self.getp('legendlabel')[idx] = value

    def get_legendlabel_prop(self, a):
        return self.getp('legendlabelprop')

    def set_legendlabel_prop(self, v, a):
        #        text = self._get_texts()[0]
        for text in a.texts + [self._get_title()]:
            page = a.figure.figobj
            if v[0] != 'default':
                text.set_family(v[0])
            else:
                text.set_family(page.getp('tick_font'))
            if v[1] != 'default':
                text.set_weight(v[1])
            else:
                text.set_weight(page.getp('tick_weight'))
            if v[2] != 'default':
                text.set_style(v[2])
            else:
                text.set_style(page.getp('tick_style'))
            if v[3] != 'default':
                text.set_size(v[3])
            else:
                text.set_size(page.getp('ticklabel_size'))

        self.setp('legendlabelprop', v)

    def get_legendentry(self, a):
        labels = [t.get_text() for t in self._get_texts()]
        colors = [self.getp('legendlabelcolor')[idx]
                  for idx, t in enumerate(self._get_texts())]

        idxlabel = ['#'+str(x+1) for x in range(len(labels))]
        if self._hit_a is not None and self._hit_a in self._artists[0].get_texts():
            idx = self._artists[0].get_texts().index(self._hit_a)
        else:
            idx = -1
        return idx, idxlabel, list(zip(labels, colors))

    def set_legendentry(self, value, a):
        idx = value[0]
        v = value[2][idx]

        self.getp('legendlabelcolor')[idx] = v[1]

        text = self._get_texts()[idx]
        text.set_text(v[0])
        text.set_color(v[1])
        h = self._artists[0].legendHandles[idx]
        h.set_label(v[0])
        self.getp('legendlabel')[idx] = v[0]

    def get_legendloc(self, a):
        return (self.getp('draggable'),
                self.getp('legendloc'),
                self.getp('legendpos'), )

    def set_legendloc(self, value, a):
        '''
        value = [draggable, loc, pos]
        '''

        self.setp('draggable', value[0])
        if len(value) == 2:
            self.setp('legendloc', value[1])
            if not value[0]:
                self.setp('legendpos', (None, None))
        if len(value) == 3:
            self.setp('legendpos', value[2])
            self.setp('legendloc', value[1])
        self.set_update_artist_request()

    def get_facecolor(self, a):
        return self._artists[0].legendPatch.get_facecolor()

    def set_facecolor(self, value, a):
        self._artists[0].legendPatch.set_facecolor(value)
        self.setvar('fbfacecolor', value)

    def get_edgecolor(self, a):
        return self._artists[0].legendPatch.get_edgecolor()

    def set_edgecolor(self, value, a):
        self._artists[0].legendPatch.set_edgecolor(value)
        self.setvar('fbedgecolor', value)

    def get_alpha(self, a):
        return self._artists[0].legendPatch.get_alpha()

    def set_alpha(self, value, a):
        self._artists[0].legendPatch.set_alpha(value)
        self.setvar('fbalpha', value)

    def get_linewidth(self, a):
        return self._artists[0].legendPatch.get_linewidth()

    def set_linewidth(self, value, a):
        self._artists[0].legendPatch.set_linewidth(value)
        self.setvar('fblinewidth', value)

    def get_linestyle(self, a):
        return self._artists[0].legendPatch.get_linestyle()

    def set_linestyle(self, value, a):
        self._artists[0].legendPatch.set_linestyle(value)
        self.setvar('fblinestyle', value)

    def get_title(self, a):
        if self._artists[0].get_title() is None:
            return ''
        else:
            return self._artists[0].get_title().get_text()

    def set_title(self, value, a):
        self.setvar('title', value)
        self.set_update_artist_request()

    def get_shadow(self, a):
        return self.getvar('shadow')

    def set_shadow(self, value, a):
        self.setvar('shadow', value)
        self.set_update_artist_request()

    def double_click_on_canvas(self, event, canvas):
        import matplotlib
        if not isinstance(self._hit_a, matplotlib.text.Text):
            return
        target_artist = self._hit_a
        if self._hit_a in self._get_texts():
            idx = self._get_texts().index(self._hit_a)
        elif self._hit_a is self._get_title():
            idx = -1
        else:
            return
#            idx = self._get_texts().index(self._hit_a)
#        if idx == -1: return
#        print idx
        tab = 'label1' if idx == 0 else str(idx+1)
        current_txt = target_artist.get_text()
        y = abs(event.y - canvas.GetClientSize()[1])

        def finish_text_edit(x, y, txt, target_artist=self._artists[0],
                             obj=canvas, tab=str(idx+1)):
            self = obj
            if tab != '0':
                a1 = UndoRedoFigobjMethod(target_artist,
                                          'legendlabel', txt)
                a1.set_extrainfo(tab)
            else:
                a1 = UndoRedoFigobjMethod(target_artist,
                                          'title', txt)
            window = self.GetTopLevelParent()
            GlobalHistory().get_history(window).make_entry([a1])
        canvas.ask_text_input(event.x, y,
                              value=current_txt, callback=finish_text_edit)

    def _tab_2_idx(self, tab):
        if tab == 'label1':
            idx = 0
        else:
            idx = int(tab)-1
        return idx

    def _get_texts(self):
        if len(self._artists) == 0:
            return []
        return self._artists[0].get_texts()

    def _get_title(self):
        return self._artists[0].get_title()

    def onRefreshLegend(self, evt):
        self.set_update_artist_request()
        canvas = evt.GetEventObject()
        ax = self.get_figaxes()
        if ax is not None:
            ax.set_bmp_update(False)
        canvas.draw()

    def canvas_menu(self):
        if self.get_rasterized():
            return [("Refresh Legend",   self.onRefreshLegend, None),
                    ("^Rasterized",  self.onUnsetRasterize, None)]
        else:
            return [("Refresh Legend",   self.onRefreshLegend, None),
                    ("*Rasterized",  self.onSetRasterize, None)]



import os
from ifigure.utils.add_sys_path import AddSysPath
from ifigure.widgets.dlg_preference import PrefComponent
from ifigure.utils.setting_parser import iFigureSettingParser as SP
import ifigure.utils.debug as debug
dprint1, dprint2, dprint3 = debug.init_dprints('HelperApp')

#
#  Plan is...
#      choice of button setting in plot
#      keyboard binding rule
#


class AppearanceConfig(PrefComponent):
    def __init__(self):
        PrefComponent.__init__(self, 'Appearance')
        p = SP()
        self.setting = p.read_setting('pref.appearance_config')

    def save_setting(self):
        p = SP()
        p.write_setting('pref.appearance_config', self.setting)

    def get_dialoglist(self):
        list1 = [["Suppress/Delete Menu in TreeViewer", self.setting["show_suppress_menu"], 3, {"text": "Show"}],
                 ["Use 3D graphics (needs restart)",
                  self.setting["gl_use"], 3, {"text": "On"}],
                 ["OpenGL multisampling",
                     self.setting["gl_multisample"], 3, {"text": "On"}],
                 ["Use OpenGL 2.1 (instead of 3.2) (needs restart)",
                  self.setting["gl_use_12"], 3, {"text": "On"}],
                 ["Generate additional screen refresh",
                  self.setting["generate_more_refresh"], 3, {"text": "On"}],
                 ]

        hint1 = ['Show Suppress, Unsuppress, and Delete Menu in Project Tree Viewer',
                 'Use super reslution for smooth 3D graphics',
                 'Turn on OpenGL 3D graphics',
                 'OpenGL 3.2 runs only on newer computers',
                 'Refresh GUI more often for some remote client.',
                 ]

        return list1, hint1

    def set_dialog_result(self, value):
        self.setting["show_suppress_menu"] = value[0]
        self.setting["gl_use"] = value[1]
        self.setting["gl_multisample"] = value[2]
        self.setting["gl_use_12"] = value[3]
        self.setting["generate_more_refresh"] = value[4]
        import ifigure.matplotlib_mod.backend_wxagg_gl
        ifigure.matplotlib_mod.backend_wxagg_gl.multisample_init_done = False



import wx
import os
import time
import random
########################################################################
#from ifigure.widgets.primitive_widgets import primitive_widgets
from ifigure.widgets.section_editor import section_editor
from ifigure.widgets.artist_widgets import panel1, panel2

from ifigure.utils.wx3to4 import GridSizer, FlexGridSizer


class property_editor(wx.Panel):
    screen_width = None
#    screen_width = 290

    def __init__(self, parent=None):
        super(property_editor, self).__init__(parent)
        self._last_update = 0
        self.ifigure_canvas = None

        self.b1 = wx.ToggleButton(self, label='artist')
        self.b2 = wx.ToggleButton(self, label='axes')
        self.b3 = wx.ToggleButton(self, label='page')
        self.b1.Bind(wx.EVT_TOGGLEBUTTON, self.ToggleArtist)
        self.b2.Bind(wx.EVT_TOGGLEBUTTON, self.ToggleAxes)
        self.b3.Bind(wx.EVT_TOGGLEBUTTON, self.ToggleSection)

        if (property_editor.screen_width == None
                or property_editor.screen_width == 0):
            self.CP1 = panel1(self, False)
        else:
            self.CP1 = panel1(self, True)
        self.CP2 = panel2(self)
        self.CP3 = section_editor(self)
#        self.Bind(self.CP1.events, self.hndl_event)
#        self.Bind(self.CP2.events, self.hndl_event)
#        self.Bind(self.CP3.events, self.hndl_event)

        sizer = wx.BoxSizer(wx.VERTICAL)
        button_sizer = GridSizer(1, 3)
        button_sizer.Add(self.b1, 0)
        button_sizer.Add(self.b2, 0)
        button_sizer.Add(self.b3, 0)
        sizer.Add(button_sizer, 0, wx.ALIGN_CENTER)

        sizer.Add(self.CP1, 1, wx.EXPAND | wx.ALL, 3)
#        sizer.Add(self.CP2, 1, wx.EXPAND|wx.ALL, 3)
#        sizer.Add(self.CP3, 1, wx.EXPAND|wx.ALL, 3)
        self.sizer_olditem = self.CP1

        self.SetSizer(sizer)
#        self.SetSizeHints(250, -1, maxW=250)
        self.Layout()
#        self.parent = parent
        self.CP1.GetSizer().Layout()

        # initial setting
        self.b1.SetValue(True)
        self.b2.SetValue(False)
        self.b3.SetValue(False)
        self.CP2.Show()
        self.CP1.Hide()
        self.CP3.Hide()
        self.OnPaneChanged()

        # this flags record if panels are ever shown on Screen.
        # if not, GTK3 shows
        #  "lost focus even though it didn't have it"
        # Debug messages.
        self._panel_shown_flag = [True, False, False]

    @property
    def parent(self):
        return self.GetParent()

    def set_sizehint(self):
        if (property_editor.screen_width == None
                or property_editor.screen_width == 0):
            # print 'size scan start'
            #        if True:
            s = self.IsShown()

            self.Show()
            m = -1
            self.CP1.Hide()
            self.CP2.Show()
            self.CP3.Hide()
            # print 'cp2'
            self.GetSizer().Replace(self.sizer_olditem, self.CP2)
            self.sizer_olditem = self.CP2
            for p in self.CP2.panels:
                self.CP2.switch_panel(p)
                self.Layout()
                self.CP2.GetSizer().Layout()
                self.OnPaneChanged()
                # print p, self.CP2.panels[p].GetSize()
                m = max([m, self.CP2.panels[p].GetSize()[0]])
            self.CP1.Show()
            self.CP2.Hide()
            self.CP3.Hide()
            # print 'cp1'
            self.GetSizer().Replace(self.sizer_olditem, self.CP1)
            self.sizer_olditem = self.CP1
            for p in self.CP1.panels:
                self.CP1.switch_panel(p)
                self.Layout()
                self.CP1.GetSizer().Layout()
                self.OnPaneChanged()
                self.CP1.Layout()
                # print p, self.CP1.panels[p].GetSize()
                m = max([m, self.CP1.panels[p].GetSize()[0]])

            self.CP1.switch_panel('text')
            self.Layout()
            self.CP1.GetSizer().Layout()
            self.OnPaneChanged()
            property_editor.screen_width = m
            # print 'panel size ', m
            self.Show(s)
        else:
            m = property_editor.screen_width
            self.CP1.Show()
            self.CP2.Hide()
            self.CP3.Hide()
            self.CP1.switch_panel('text')
            self.CP2.switch_panel('axcommon.x.y')
            self.Layout()
            self.CP1.GetSizer().Layout()
            self.CP2.GetSizer().Layout()
            self.OnPaneChanged()
        if m > 10:
            self.SetSizeHints(m, -1, maxW=m)
            for pname in self.CP1.panels:
                self.CP1.panels[pname].SetSizeHints(m, -1, maxW=m)
                #self.CP1.panels[pname].SetSize((m, -1))
#            for pname in  self.CP2.panels:
#                for elp in self.CP2.panels[pname].elp:
#                    elp.EnableScrolling(False, True)
#                    elp.SetScrollRate(0,5)
#                    elp.SetMinVirtualSize(m)

    def set_canvas(self, canvas):
        self.ifigure_canvas = canvas

    def get_canvas(self):
        return self.ifigure_canvas

    def ToggleArtist(self, e):
        obj = e.GetEventObject()
        isPressed = obj.GetValue()
        sizer = self.GetSizer()
        if isPressed:
            self.b2.SetValue(False)
            self.b3.SetValue(False)
            self.CP1.Show()
            self.CP2.Hide()
            self.CP3.Hide()
            if self.sizer_olditem is not self.CP1:
                sizer.Replace(self.sizer_olditem, self.CP1)
                self.sizer_olditem = self.CP1
        else:
            self.b1.SetValue(True)
        self.CP1.GetSizer().Layout()
        self.CP1.update_panel()
        self.OnPaneChanged()
        self._panel_shown_flag[0] = True

        tw = wx.GetApp().TopWindow
        if tw.appearanceconfig.setting['generate_more_refresh']:
            self.GetTopLevelParent().deffered_force_layout()

    def ToggleAxes(self, e):
        obj = e.GetEventObject()
        isPressed = obj.GetValue()
        sizer = self.GetSizer()
        if isPressed:
            self.b1.SetValue(False)
            self.b3.SetValue(False)
            self.CP1.Hide()
            self.CP3.Hide()
            if self.sizer_olditem is not self.CP2:
                sizer.Replace(self.sizer_olditem, self.CP2)
                self.sizer_olditem = self.CP2
            self.CP2.Layout()
            self.CP2.Show()
        else:
            self.b2.SetValue(True)
        self.CP2.GetSizer().Layout()

        # set selected axes to panel
        canvas = self.get_canvas()
        ax = canvas.axes_selection()
        self.CP2.set_axes(ax)

        self.CP2.update_panel()
        self.OnPaneChanged()
        self.CP2.Layout()
        self._panel_shown_flag[1] = True

        tw = wx.GetApp().TopWindow
        if tw.appearanceconfig.setting['generate_more_refresh']:
            self.GetTopLevelParent().deffered_force_layout()

    def ToggleSection(self, e):
        obj = e.GetEventObject()
        isPressed = obj.GetValue()
        sizer = self.GetSizer()
        if isPressed:
            self.b1.SetValue(False)
            self.b2.SetValue(False)
            self.CP3.Show()
            self.CP1.Hide()
            self.CP2.Hide()
            if self.sizer_olditem is not self.CP3:
                sizer.Replace(self.sizer_olditem, self.CP3)
                self.sizer_olditem = self.CP3
        else:
            self.b3.SetValue(True)
        self.CP3.SetEditorValue()
        self.CP3.GetSizer().Layout()
        self.CP3.update_panel()
        self.OnPaneChanged()
        self._panel_shown_flag[2] = True

        tw = wx.GetApp().TopWindow
        if tw.appearanceconfig.setting['generate_more_refresh']:
            self.GetTopLevelParent().deffered_force_layout()

    def hndl_event(self, event):
        if event.GetEventObject() is self.CP1:
            pass
        if event.GetEventObject() is self.CP2:
            self.ifigure_canvas.draw()
        if event.GetEventObject() is self.CP3:
            pass

    def SetCanvasValue(self, message=None):
        self.CP1.SetCanvasValue(message)

        #  Update Axes Panel
        # self.CP2.SetCanvasValue(message)

        #  Update Section Panel
        self.CP3.SetCanvasValue(message)

    def update_panel_request2(self, time):
        if (time > self._last_update):
            self.update_panel()

    def update_panel(self):
        if self.IsShown():
            if self.CP1.IsShown():
                self.CP1.update_panel()
#                if len(self.get_canvas().selection) > 0:

            if self.CP2.IsShown():
                self.CP2.update_panel()
            if self.CP3.IsShown():
                self.CP3.update_panel()
            self._last_update = time.time()
        self.Layout()

    def OnPaneChanged(self):
        # redo the layout
        #        self.GetSizer().Layout()
        #   self.Fit()
        self.GetParent().Layout()
    #   self.parent.Fit()

    def onTD_Selection(self, evt):
        if self.IsShown():
            if self._panel_shown_flag[0]:
                self.CP1.onTD_Selection(evt)
            if self._panel_shown_flag[1]:
                self.CP2.onTD_Selection(evt)
            if self._panel_shown_flag[2]:
                self.CP3.onTD_Selection(evt)
        self.Layout()
        # this will make sure that button size is right..!?

    def onTD_ShowPage(self, evt):
        if self.IsShown():
            self.CP1.update_panel()
            self.CP2.update_panel()
            self.CP3.onTD_ShowPage(evt)

    def onTD_Replace(self, evt):
        if self.IsShown():
            self.CP1.onTD_Replace(evt)
            self.CP2.onTD_Replace(evt)
            self.CP3.onTD_Replace(evt)

    def save_screencapture(self, file='~/test.png'):
        import os
        file = os.path.expanduser(file)
        context = wx.ClientDC(self)
        memory = wx.MemoryDC()
        x, y = self.ClientSize
        bitmap = wx.EmptyBitmap(x, y, -1)
        memory.SelectObject(bitmap)
        memory.Blit(0, 0, x, y, context, 0, 0)
        memory.SelectObject(wx.NullBitmap)
        bitmap.SaveFile(file, wx.BITMAP_TYPE_PNG)



import multiprocessing as mp
import time
import numpy
import tempfile
import traceback
import threading

from ifigure.mdsplus.proxy_jobrunner import JobRunner
#from ifigure.mdsplus.fake_jobrunner import JobRunner
from numpy import ndarray, isnan, squeeze
from ifigure.mdsplus.filed_ndarray import FiledNDArray


class ProxyWorkerBase(object):
    use_compress = True

    def run_loop(self, *args, **kargs):
        runner = JobRunner(self.host, self.port)
        while True:
            try:
                task = self.task_queue.get(True)
            except EOFError:
                r = {'error message': ['message was received']}
                r['worker output'] = ''
                self.result_queue.put((-1, 0, r))
                self.task_queue.task_done()

            if task != -1:
                job_id, jobgroup, globaljob = task
                ret = [None]*len(jobgroup)
                for k, jobset in enumerate(jobgroup):
                    skip_init = False
                    jobs, names = jobset
                    for job in jobs:
                        for j in job:
                            if j.command == 'open':
                                shot = j.params[1]
                                if shot == runner._shot:
                                    skip_init = True
                                    break
                runner.set_globaljob(globaljob)
                for k, jobset in enumerate(jobgroup):
                    flag = False
                    try:
                        r = runner.run(jobset, skip_init)
                    except:
                        r = {'error message': [
                            'runner.run failed', traceback.format_exc()]}
                        flag = True
                    finally:
                        r['worker output'] = ' '  # tmp.read()
                        #r['error message'] = []
                    try:
                        jobs, names = jobset
                        can_compress = jobs.can_compress
                        def_names = ['message',
                                     'pickle error',
                                     'network message',
                                     'worker output',
                                     'error message',
                                     'x', 'y', 'z',
                                     'xerr', 'yerr', 'zerr']
                        for n in names:
                            if not n in def_names:
                                def_names.append(n)
                        r2 = {}
                        for n in def_names:
                            if n in r:
                                if isinstance(r[n], ndarray) and can_compress:
                                    r[n] = squeeze(r[n])
                                if (isinstance(r[n], ndarray) and
                                    r[n].size > 5e4 and len(r[n].shape) == 1 and
                                    can_compress and not any(isnan(r[n])) and
                                        ProxyWorkerBase.use_compress):
                                    r2[n] = FiledNDArray(r[n])
                                else:
                                    #                                 r2[n] = FileNDArray(r[n])
                                    r2[n] = r[n]
                        ret[k] = r2
                        r = r2
                    except:
                        r['error message'].append(traceback.format_exc())
                        ret[k] = r
                    self.result_queue.put((job_id, k, r))
#                   if flag: break
#                print 'putting data in queue', time.time()
#                self.result_queue.put((job_id, ret))
                self.task_queue.task_done()
            else:
                break
        runner.terminate()
        self.task_queue.task_done()
#                self.result_queue.put('ok')


class MDSMPProxyWorker(mp.Process, ProxyWorkerBase):
    def __init__(self, task_queue, result_queue, host, port, *args, **kargs):
        #   def __init__(self,  *args, **kargs):
        super(MDSMPProxyWorker, self).__init__(*args, **kargs)
        self.task_queue = task_queue
        self.result_queue = result_queue
        self.port = int(port)
        self.host = host

    def run(self, *args, **kargs):
        return ProxyWorkerBase.run_loop(self, *args, **kargs)


class MDSTHProxyWorker(threading.Thread, ProxyWorkerBase):
    def __init__(self, task_queue, result_queue, host, port, *args, **kargs):
        #   def __init__(self,  *args, **kargs):
        super(MDSTHProxyWorker, self).__init__(*args, **kargs)
        self.task_queue = task_queue
        self.result_queue = result_queue
        self.port = int(port)
        self.host = host
        ProxyWorkerBase.use_compress = True

    def run(self, *args, **kargs):
        return ProxyWorkerBase.run_loop(self, *args, **kargs)
